/*
 ============================================================================
 Name		 : MasteroidsEngine.cpp
 Copyright   : Symbian Academy, 2007
 Description : This engine handles the game relevant implementation.
 ============================================================================
 */

#ifndef MASTEROIDSENGINE_H
#define MASTEROIDSENGINE_H

// INCLUDES
#include <e32std.h>
#include <e32base.h>
#include <w32std.h>
#include <coedef.h>
#include <s32mem.h>
#include "MSocketsNotifier.h"
#include "MMessageDisplay.h"
#include "SocketsEngine.h"

#include <e32debug.h>

// CLASS DECLARATION

/**
 *  CMasteroidsEngine
 * 
 */
class CMasteroidsEngine : public CBase, public MSocketsNotifier
	{
public:
	// Constructors and destructor

	/**
	 * Destructor.
	 */
	~CMasteroidsEngine();

	/**
	 * Two-phased constructor.
	 */
	static CMasteroidsEngine* NewL();

	/**
	 * Two-phased constructor.
	 */
	static CMasteroidsEngine* NewLC();

private:

	/**
	 * Constructor for performing 1st stage construction
	 */
	CMasteroidsEngine();

	/**
	 * EPOC default constructor for performing 2nd stage construction
	 */
	void ConstructL();

public:
	/**
	 * Store a pointer to the sockets engine, so that the game engine
	 * can send out requests to the masteroids server.
	 * 
	 * @param aSocketsEngine pointer to the sockets engine.
	 */
	void SetSocketsEngine(CSocketsEngine* aSocketsEngine);

	/**
	 * Store a pointer to the user interface, which is used
	 * to display information or error messages to the user.
	 * 
	 * @param aMessageDisplay the message to show to the user.
	 */
	void SetMessageDisplay(MMessageDisplay* aMessageDisplay);

	/**
	 * Process the key event.
	 * 
	 * @param aKeyEvent contains the key code.
	 * @param aType contains information about whether the button was pressed
	 * or released.
	 */
	TKeyResponse ProcessKeyEventL(const TKeyEvent& aKeyEvent, TEventCode aType);

	/**
	 * Function defined by the MSocketsNotifier interface.
	 * Allows the socket engine to send information / error messages to the
	 * game engine.
	 * 
	 * @param aError error code, KErrNone (0) if it's an information
	 * message.
	 * @param aMsg message text.
	 */
	void NotifyInfo(TInt aError, const TDesC& aMsg);

	/**
	 * Function defined by the MSocketsNotifier interface.
	 * Whenever the sockets reader receives data, it is sent
	 * to this class through this method.
	 * 
	 * @param aData 8 bit descriptor that contains the data
	 * received through the socket reader.
	 */
	void MessageReceived(const TDesC8& aData);

	/**
	 * Method that can be called from the user interface to send
	 * the login information to the server.
	 * 
	 * @param aUsername 8 bit descriptor containing the username
	 * of this client.
	 */
	void SendLoginL(const TDesC8& aUsername);

	/**
	 * Query if the game is currently logged in.
	 * 
	 * @return ETrue if a successful login request has already
	 * been executed, EFalse otherwise.
	 */
	TBool IsLoggedIn();

	/**
	 * Query if a game is currently active.
	 * 
	 * @return ETrue if a game is running at the moment.
	 */
	TBool IsGameStarted();
private:
	/**
	 * Prepare and send a control message containing control updates to the
	 * server, according to the protocol.
	 * 
	 * @param aAction control action, according to the protocol.
	 * @param aParameter additional parameters, according to the protocol
	 * and the executed action.
	 */
	void SendControlMessageL(TUint8 aAction, TUint8 aParameter);

	/**
	 * Parse messages from the incoming data.
	 * This method combines new parts of data and individually
	 * seperates messages if more than one is in the stream.
	 * 
	 * @param aData the data to add to the internal buffer,
	 * that will be parsed by this method.
	 */
	void ParseMessageL(const TDesC8& aData);

	/**
	 * Process an individual message according to the protocol.
	 * If more than one message is contained in the descriptor,
	 * it has to be parsed into two seperate descriptors by
	 * the ParseMessageL() method first.
	 * 
	 * @param aData descriptor that contains only one game message.
	 */
	void ProcessMessageL(const TDesC8& aData);

	/**
	 * Prepare and send a rotation message to the server.
	 * 
	 * @param aRotationMode 0 = stop rotating,
	 * 1 = start rotating left,
	 * -1 / 255 = start rotating right.
	 */
	void SendRotationL(TInt aRotationMode);

	/**
	 * Start or stop accelerating.
	 * 
	 * @param aThrustFlag ETrue = start accelerating,
	 * EFalse = stop accelerating.
	 */
	void SendThrustL(TBool aThrustFlag);

	/**
	 * Send a single fire event.
	 */
	void SendFireL();
private:
	/**
	 * Whether a (successful) login request has been done.
	 */
	TBool iLoggedIn;

	/**
	 * Indicates if a game is currently running.
	 */
	TBool iGameStarted;

	/**
	 * Buffer for combining data received from the sockets
	 * engine, in case a single message is split over more
	 * than one RecvOneOrMore()-events.
	 */
	TBuf8<300> iMsgBuffer;

	/**
	 * Reference to the socket engine for sending commands
	 * to the game server.
	 */
	CSocketsEngine* iSocketsEngine; // owned by the AppUi

	/**
	 * Reference to an UI interface that can display information
	 * and error messages.
	 */
	MMessageDisplay* iMessageDisplay;

	// Various message ids for the game protocol:
	static const TUint8 MID_LOGIN_REQUEST = 1;
	static const TUint8 MID_LOGIN_RESPONSE = 2;
	static const TUint8 MID_PHYSICS = 3;
	static const TUint8 MID_GO_STOP = 4;
	static const TUint8 MID_CONTROL = 5;
	static const TUint8 MID_PLAYERS_INFO = 6;
	static const TUint8 MID_BULLETS_INFO = 7;
	};

#endif // MASTEROIDSENGINE_H

