/*
 ============================================================================
 Name		 : MasteroidsEngine.cpp
 Copyright   : Symbian Academy, 2007
 Description : This engine handles the game relevant implementation.
 ============================================================================
 */

#include "MasteroidsEngine.h"

CMasteroidsEngine::CMasteroidsEngine() :
	iLoggedIn(EFalse), iGameStarted(EFalse)
	{
	// No implementation required
	}

CMasteroidsEngine::~CMasteroidsEngine()
	{
	}

CMasteroidsEngine* CMasteroidsEngine::NewLC()
	{
	CMasteroidsEngine* self = new (ELeave)CMasteroidsEngine();
	CleanupStack::PushL (self);
	self->ConstructL ();
	return self;
	}

CMasteroidsEngine* CMasteroidsEngine::NewL()
	{
	CMasteroidsEngine* self=CMasteroidsEngine::NewLC ();
	CleanupStack::Pop (); // self;
	return self;
	}

void CMasteroidsEngine::ConstructL()
	{

	}

void CMasteroidsEngine::SetSocketsEngine(CSocketsEngine* aSocketsEngine)
	{
	iSocketsEngine = aSocketsEngine;
	}

void CMasteroidsEngine::SetMessageDisplay(MMessageDisplay* aMessageDisplay)
	{
	iMessageDisplay = aMessageDisplay;
	}

TBool CMasteroidsEngine::IsLoggedIn()
	{
	return iLoggedIn;
	}

TBool CMasteroidsEngine::IsGameStarted()
	{
	return iGameStarted;
	}

TKeyResponse CMasteroidsEngine::ProcessKeyEventL(const TKeyEvent& aKeyEvent,
		TEventCode aType)
	{
	// If no game is currently active, don't send key events to the server.
	if ( !IsGameStarted())
		return EKeyWasNotConsumed;

	TKeyResponse response = EKeyWasConsumed;
	// Check if the key was pressed or released - depending on this, 
	// adapt the behavior of the engine.
	if ( aType == EEventKeyDown) // Key was pressed
		{
		switch (aKeyEvent.iScanCode)
			{
			case EStdKeyUpArrow: // Thrust
				SendThrustL (ETrue);
				break;
			case EStdKeyLeftArrow: // Rotate left
				SendRotationL (1);
				break;
			case EStdKeyRightArrow: // Rotate right
				SendRotationL (-1);
				break;
			case EStdKeyDevice3: // Fire
			case EStdKeyEnter:
			case EStdKeyNkpEnter:
			case EStdKeySpace:
			case EKeySpace:
			case EKeyEnter:
				SendFireL ();
				break;
			default:
				response = EKeyWasNotConsumed;
				break;
			}
		}
	else
		if ( aType == EEventKeyUp) // Key was released
			{
			switch (aKeyEvent.iScanCode)
				{
				case EStdKeyUpArrow: // Stop accelerating
					SendThrustL (EFalse);
					break;
				case EStdKeyLeftArrow: // Stop rotating
					SendRotationL (0);
					break;
				case EStdKeyRightArrow: // Stop rotating
					SendRotationL (0);
					break;
				default:
					response = EKeyWasNotConsumed;
					break;
				}
			}
	return response;
	}

void CMasteroidsEngine::NotifyInfo(TInt aError, const TDesC& aMsg)
	{
	if ( iMessageDisplay)
		{
		// Format the notification message to contain the error code as
		// well as the message
		_LIT(KErrorMsg, "%d: %S");
		TBuf<100> tmpMsg;
		tmpMsg.Format (KErrorMsg, aError, &aMsg);
		TRAPD(err, iMessageDisplay->SetStatusTextL(tmpMsg))
		;
		}
	}

void CMasteroidsEngine::MessageReceived(const TDesC8& aData)
	{
	// Handle the message and trap any errors in this central place
	// to prevent multiple TRAPs
	TRAPD(err, ParseMessageL(aData))
	;
	// Any possible errors should be handled here to make
	// the game more robust.
	}

void CMasteroidsEngine::ParseMessageL(const TDesC8& aData)
	{
	// Add the new data to the message buffer
	iMsgBuffer.Append (aData);
	// Find out if the message buffer is already long enough to contain
	// at least one message (header length is 4, so the message should be
	// at least 4 bytes long)
	while (iMsgBuffer.Length ()>= 4)
		{
		// Read the message length from the 3rd and 4th byte in little endian
		TUint16 msgLength = (TUint16)(iMsgBuffer[2] | iMsgBuffer[3] << 8);
		// Check if the buffer already contains the header (4 bytes) + the 
		// whole message, whose length is sent through the protocol
		if ( iMsgBuffer.Length ()< msgLength + 4)
			{
			// Data isn't complete yet...
			return;
			}
		// At least one message is ready!
		// Prepare a pointer descriptor that only contains this message
		// from the message buffer.
		TPtrC8 curData = iMsgBuffer.Left (msgLength + 4);
		// Process the message
		ProcessMessageL (curData);
		// Delete this message from the message buffer
		iMsgBuffer.Delete (0, msgLength + 4);
		// ... if the length is still more than 4, the loop will
		// be processed again. There might be 2+ messages in the buffer.
		}
	}

void CMasteroidsEngine::ProcessMessageL(const TDesC8& aData)
	{
	// Create a read stream for easier access to the data
	RDesReadStream readStream(aData);
	// The stream has to be closed!
	CleanupClosePushL (readStream);
	// Read message id
	TUint16 msgId = readStream.ReadUint16L ();
	// Read message length
	TUint16 msgLength = readStream.ReadUint16L ();
	if ( msgLength + 4 != aData.Length ())
		{
		// Error: Wrong message length in package
		CleanupStack::PopAndDestroy (1);
		return;
		}
	// Start the right action depending on the message id
	// Note: Not all messages are handled by this implementation.
	// Reading the other messages would be necessary for implementing
	// a local display or an AI.
	switch (msgId)
		{
		case MID_LOGIN_RESPONSE:
			{
			// If data is 1, we are logged in.
			// If it's 0, the login was not successful.
			iLoggedIn = (readStream.ReadUint8L() == 1);
			_LIT(KLoggedIn, "Logged in: %d");
			TBuf<100> tmpMsg;
			tmpMsg.Format (KLoggedIn, iLoggedIn);
			iMessageDisplay->SetStatusTextL (tmpMsg);
			}
			break;
		case MID_PHYSICS:
			break;
		case MID_GO_STOP:
			{
			// If data is 1, the game has started.
			// If it's 0, the game was ended.
			iGameStarted = (readStream.ReadUint8L() == 1);
			_LIT(KGameStarted, "Game started: %d");
			TBuf<100> tmpMsg;
			tmpMsg.Format (KGameStarted, iGameStarted);
			iMessageDisplay->SetStatusTextL (tmpMsg);
			}
			break;
		case MID_PLAYERS_INFO:
			break;
		case MID_BULLETS_INFO:
			break;
		}
	CleanupStack::PopAndDestroy ();
	}

void CMasteroidsEngine::SendLoginL(const TDesC8& aUsername)
	{
	// Allocate the memory for the message
	RBuf8 tmpMsg;
	tmpMsg.CreateL (aUsername.Length ()+ 4);
	CleanupClosePushL (tmpMsg);
	// Create a write stream for the message data
	RDesWriteStream stream(tmpMsg);
	CleanupClosePushL (stream);
	stream.WriteUint16L (MID_LOGIN_REQUEST);
	stream.WriteUint16L (aUsername.Length ());
	stream.WriteL (aUsername);
	// Don't forget to commit the data to the descriptor,
	// otherwise it wouldn't work!
	stream.CommitL ();
	// Send the prepared message
	iSocketsEngine->IssueSend (tmpMsg);
	CleanupStack::PopAndDestroy (2);
	}

void CMasteroidsEngine::SendRotationL(TInt aRotationMode)
	{
	SendControlMessageL (0, (TUint8)aRotationMode);
	}

void CMasteroidsEngine::SendThrustL(TBool aThrustFlag)
	{
	SendControlMessageL (1, (TUint8)(aThrustFlag ? 1 : 0));
	}

void CMasteroidsEngine::SendFireL()
	{
	SendControlMessageL (2, 0);
	}

void CMasteroidsEngine::SendControlMessageL(TUint8 aAction, TUint8 aParameter)
	{
	// Help function to put the provided parameters in a message that is
	// compiled according to the game protocol.
	TBuf8<6> tmpMsg;
	RDesWriteStream stream(tmpMsg);
	CleanupClosePushL (stream);
	stream.WriteUint16L (MID_CONTROL);
	stream.WriteUint16L (2);
	stream.WriteInt8L (aAction);
	stream.WriteInt8L (aParameter);
	stream.CommitL ();
	iSocketsEngine->IssueSend (tmpMsg);
	CleanupStack::PopAndDestroy (1);
	}
